from elastic.algorithm.selector import Selector
from elastic.core.graph.variable_snapshot import VariableSnapshot
from elastic.core.graph.cell_execution import CellExecution

from collections import defaultdict
import networkx as nx
from networkx.algorithms.flow import shortest_augmenting_path
import numpy as np


class OptimizerExact(Selector):
    """
        The exact optimizer constructs a flow graph and runs the min-cut algorithm to exactly find the best
        checkpointing configuration.
    """

    def __init__(self, migration_speed_bps=1):
        super().__init__(migration_speed_bps)

        # Augmented computation graph
        self.active_oes = None
        self.compute_graph = None

        # CEs required to recompute a variables last modified by a given CE.
        self.recomputation_ces = {}

        self.idx = 0

    def get_new_idx(self) -> int:
        """
            Get a new index number to add to compute graph.
        """
        idx = self.idx
        self.idx += 1
        return idx

    def dfs(self, current: str, visited: set, recompute_ces: str):
        """
            Perform DFS on the Application History Graph for finding the CEs required to recompute a variable.
            Args:
                current (str): Name of current nodeset.
                visited (set): Visited nodesets.
                recompute_ces (set): Set of CEs needing re-execution to recompute the current nodeset.
        """
        if isinstance(current, CellExecution):
            # Result is memoized
            if current in self.recomputation_ces:
                recompute_ces.update(self.recomputation_ces[current])
            else:
                recompute_ces.add(current)
                for vs in current.src_vss:
                    if vs not in self.active_vss and vs not in visited:
                        self.dfs(vs, visited, recompute_ces)

        elif isinstance(current, VariableSnapshot):
            visited.add(current)
            if current.output_ce not in recompute_ces:
                self.dfs(current.output_ce, visited, recompute_ces)

    def find_prerequisites(self):
        """
            Find the necessary (prerequisite) cell executions to rerun a cell execution.
        """
        self.active_vss = set(self.active_vss)

        for ce in self.dependency_graph.cell_executions:
            if ce.dst_vss.intersection(self.active_vss):
                recompute_ces = set()
                self.dfs(ce, set(), recompute_ces)
                self.recomputation_ces[ce] = recompute_ces

    def select_vss(self, write_log_location=None, notebook_name=None, optimizer_name=None) -> set:
        self.find_prerequisites()

        # Construct flow graph for computing mincut
        mincut_graph = nx.DiGraph()

        # Add source and sink to flow graph.
        mincut_graph.add_node("source")
        mincut_graph.add_node("sink")

        # Add all active VSs as nodes, connect them with the source with edge capacity equal to migration cost.
        for active_vs in self.active_vss:
            mincut_graph.add_node(active_vs)
            mincut_graph.add_edge("source", active_vs, capacity=active_vs.size / self.migration_speed_bps)

        # Add all CEs as nodes, connect them with the sink with edge capacity equal to recomputation cost.
        for ce in self.dependency_graph.cell_executions:
            mincut_graph.add_node(ce)
            mincut_graph.add_edge(ce, "sink", capacity=ce.cell_runtime)

        # Connect each CE with its output variables and its prerequisite OEs.
        for active_vs in self.active_vss:
            for ce in self.recomputation_ces[active_vs.output_ce]:
                mincut_graph.add_edge(active_vs, ce, capacity=np.inf)

        # Add constraints: overlapping variables must either be migrated or recomputed together.
        for vs_pair in self.overlapping_vss:
            mincut_graph.add_edge(vs_pair[0], vs_pair[1], capacity=np.inf)
            mincut_graph.add_edge(vs_pair[1], vs_pair[0], capacity=np.inf)

        # Prune CEs which produce no active variables to speedup computation.
        for ce in self.dependency_graph.cell_executions:
            if mincut_graph.in_degree(ce) == 0:
                mincut_graph.remove_node(ce)

        # Solve min-cut with Ford-Fulkerson.
        cut_value, partition = nx.minimum_cut(mincut_graph, "source", "sink", flow_func=shortest_augmenting_path)

        # Determine the replication plan from the partition.
        vss_to_migrate = set(partition[1]).intersection(self.active_vss)
        ces_to_recompute = set(partition[0]).intersection(self.dependency_graph.cell_executions)
         
        return vss_to_migrate, ces_to_recompute



from inspect import isclass
from types import ModuleType
from types import FunctionType
import dill
import pandas as pd
# import polars as pl
import lightgbm
import numpy as np
import scipy
import torch
import copy
import io
import xxhash

BASE_TYPES = [type(None), FunctionType]

class ImmutableObj:
    def __init__(self):
        pass

    def __eq__(self, other):
        if isinstance(other, ImmutableObj):
            return True
        return False


# Object representing none.
class NoneObj:
    def __init__(self):
        pass

    def __eq__(self, other):
        if isinstance(other, NoneObj):
            return True
        return False

# Object representing a dataframe.
class DataframeObj:
    def __init__(self):
        pass

    def __eq__(self, other):
        if isinstance(other, DataframeObj):
            return True
        return False

class NxGraphObj:
    def __init__(self, graph):
        self.graph = graph

    def __eq__(self, other):
        if isinstance(other, NxGraphObj):
            return nx.graphs_equal(self.graph, other.graph)
        return False

class NpArrayObj:
    def __init__(self, arraystr):
        self.arraystr = arraystr
        pass

    def __eq__(self, other):
        if isinstance(other, NpArrayObj):
            return self.arraystr == other.arraystr
        return False

class ScipyArrayObj:
    def __init__(self, arraystr):
        self.arraystr = arraystr
        pass

    def __eq__(self, other):
        if isinstance(other, ScipyArrayObj):
            return self.arraystr == other.arraystr
        return False

class TorchTensorObj:
    def __init__(self, arraystr):
        self.arraystr = arraystr
        pass

    def __eq__(self, other):
        if isinstance(other, TorchTensorObj):
            return self.arraystr == other.arraystr
        return False

class ModuleObj:
    def __init__(self):
        pass

    def __eq__(self, other):
        if isinstance(other, ModuleObj):
            return True
        return False

# Object representing general unserializable class.
class UnserializableObj:
    def __init__(self):
        pass

    def __eq__(self, other):
        if isinstance(other, UnserializableObj):
            return True
        return False

class UncomparableObj:
    def __init__(self):
        pass

    def __eq__(self, other):
        if isinstance(other, UncomparableObj):
            return True
        return False

def construct_object_hash(obj, deepcopy=False):
    """
        Construct an object hash for the object. Uses deep-copy as a fallback.
    """

    if type(obj) in BASE_TYPES:
        return ImmutableObj()
    
    if isclass(obj):
        return type(obj)

    # Flag hack for Pandas dataframes: each dataframe column is a numpy array.
    # All the writeable flags of these arrays are set to false; if after cell execution, any of these flags are
    # reset to True, we assume that the dataframe has been modified.
    if isinstance(obj, pd.DataFrame):
        for (_, col) in obj.items():
            col.__array__().flags.writeable = False
        return DataframeObj()

    if isinstance(obj, pd.Series):
        obj.__array__().flags.writeable = False
        return DataframeObj()

    attr_str = getattr(obj, '__module__', None)
    if attr_str and ("matplotlib" in attr_str or "transformers" in attr_str or "networkx" in attr_str or "keras" in attr_str or "tensorflow" in attr_str):
        return UncomparableObj()

    # Object is file handle
    if isinstance(obj, io.IOBase):
        return UncomparableObj()

    if isinstance(obj, np.ndarray):
        h = xxhash.xxh3_128()
        h.update(np.ascontiguousarray(obj.data))
        str1 = h.intdigest()
        return NpArrayObj(str1)

    if isinstance(obj, scipy.sparse.csr_matrix):
        h = xxhash.xxh3_128()
        h.update(np.ascontiguousarray(obj))
        str1 = h.intdigest()
        return ScipyArrayObj(str1)

    if isinstance(obj, torch.Tensor):
        h = xxhash.xxh3_128()
        h.update(np.ascontiguousarray(obj))
        str1 = h.intdigest()
        return TorchTensorObj(str1)

    if isinstance(obj, ModuleType) or isclass(obj):
        return ModuleObj()

    # Polars dataframes are immutable.
    #if isinstance(obj, pl.DataFrame):
    #    return type(obj)

    # LightGBM dataframes are immutable.
    if isinstance(obj, lightgbm.Dataset):
        return type(obj)

    # Try to hash the object; if the object is unhashable, use deepcopy as fallback.
    try:
        h = xxhash.xxh3_128()
        h.update(obj)
        return h.intdigest()
    except:
        try:
            if deepcopy:
                return copy.deepcopy(obj)
            else:
                return obj
        except:
            # If object is not even deepcopy-able, mark it as unserializable and assume modified-on-write.
            return UnserializableObj()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2021-2022 University of Illinois
from typing import List


class CellExecution:
    """
        A cell execution (object) corresponds to a cell execution (action, i.e. press play) in the notebook session.
    """
    def __init__(self, cell_num: int, cell: str, cell_runtime: float, start_time: float, src_vss: List, dst_vss: List):
        """
            Create an operation event from cell execution metrics.
            Args:
                cell_num (int): The nth cell execution of the current session.
                cell (str): Raw cell cell.
                cell_runtime (float): Cell runtime.
                start_time (time): Time of start of cell execution. Note that this is different from when the cell was
                    queued.
                src_vss (List[VariableSnapshot]): Nodeset containing input VSs of the cell execution.
                dst_vss (List[VariableSnapshot]): Nodeset containing output VSs of the cell execution.
        """
        self.cell_num = cell_num
        self.cell = cell
        self.cell_runtime = cell_runtime
        self.start_time = start_time

        self.src_vss = src_vss
        self.dst_vss = dst_vss



import pandas as pd
from elastic.core.mutation.id_graph import construct_id_graph, is_root_equals, is_structure_equals
from elastic.core.mutation.object_hash import UncomparableObj, ImmutableObj, NoneObj, NxGraphObj, \
    TorchTensorObj, ModuleObj, UnserializableObj, NpArrayObj, ScipyArrayObj, DataframeObj, \
    construct_object_hash
import time
from types import FunctionType

from collections.abc import Iterable

BASE_TYPES = [str, int, float, bool, type(None), FunctionType, ImmutableObj, UncomparableObj, NoneObj, NxGraphObj,
              TorchTensorObj, ModuleObj, UnserializableObj, NpArrayObj, ScipyArrayObj, DataframeObj]


def base_typed(obj, visited):
    """
        Recursive reflection method to convert any object property into a comparable form.
        From: https://stackoverflow.com/questions/1227121/compare-object-instances-for-equality-by-their-attributes
    """
    T = type(obj)
    from_numpy = T.__module__ == 'numpy'

    if T in BASE_TYPES or callable(obj) or (from_numpy and not isinstance(T, Iterable)):
        return obj

    visited.add(id(obj))

    if isinstance(obj, Iterable):
        return obj
    d = obj if T is dict else obj.__dict__

    comp_dict = {}
    for k, v in d.items():
        if id(v) not in visited:
            comp_dict[k] = base_typed(v, visited)

    return comp_dict


def deep_equals(*args):
    """
        Extended equality comparison which compares objects recursively by their attributes, i.e., it also works for
        certain user-defined objects with no equality (__eq__) defined.
    """
    return all(base_typed(args[0], set()) == base_typed(other, set()) for other in args[1:])


def construct_fingerprint(obj, profile_dict):
    """
        Construct a fingerprint of the object (ID graph + hash).
    """
    start = time.time()
    id_graph, id_set = construct_id_graph(obj)
    end = time.time()
    profile_dict["idgraph"] += end - start

    start = time.time()
    object_representation = construct_object_hash(obj, deepcopy=True)
    end = time.time()
    profile_dict["representation"] += end - start

    return [id_graph, id_set, object_representation]


def compare_fingerprint(fingerprint_list, new_obj, profile_dict, input_variables_id_graph_union):
    """
        Check whether an object has been changed by comparing it to its previous fingerprint.
    """
    changed = False
    overwritten = False
    uncomparable = False

    # Hack: check for pandas dataframes and series: if the flag has been flipped back on, the object has been changed.
    if isinstance(new_obj, pd.DataFrame):
        for (_, col) in new_obj.items():
            if col.__array__().flags.writeable:
                changed = True
                break

    elif isinstance(new_obj, pd.Series):
        if new_obj.__array__().flags.writeable:
            changed = True

    # ID graph check: check whether the structure of the object has changed (i.e. its relation with other objects)
    start = time.time()

    id_graph, id_set = construct_id_graph(new_obj)

    if id_set != fingerprint_list[1] or not is_structure_equals(id_graph, fingerprint_list[0]):
        # Distinguish between overwritten variables and modified variables (i.e., x = 1 vs. x[0] = 1)
        if not is_root_equals(id_graph, fingerprint_list[0]):
            overwritten = True
        changed = True
        fingerprint_list[0] = id_graph
        fingerprint_list[1] = id_set

    end = time.time()
    profile_dict["idgraph"] += end - start

    # Value check via object hash: check whether the object's value has changed
    if not changed:
        start = time.time()
        try:
            new_repr = construct_object_hash(new_obj, deepcopy=False)

            # Variable is uncomparable
            if isinstance(new_repr, UncomparableObj):
                if id_set.intersection(input_variables_id_graph_union):
                    changed = True
                uncomparable = True
                fingerprint_list[2] = UncomparableObj()
            else:
                if not deep_equals(new_repr, fingerprint_list[2]):
                    # Variable has equality defined; the variable has been modified.
                    if "__eq__" in type(new_repr).__dict__.keys() or "eq" in type(new_repr).__dict__.keys():
                        changed = True
                    else:
                        # Object is uncomparable
                        if id_set.intersection(input_variables_id_graph_union):
                            changed = True
                        uncomparable = True
                        fingerprint_list[2] = UncomparableObj()
        except:
            # Variable is uncomparable
            if id_set.intersection(input_variables_id_graph_union):
                changed = True
            uncomparable = True
            fingerprint_list[2] = UncomparableObj()

    # Update the object hash if either:
    # 1. the object has been completely overwritten
    # 2. the object has been modified, and is of a comparable type (i.e., hashable or unhashable but has equality
    # defined)
    if overwritten or (changed and not uncomparable and not isinstance(fingerprint_list[2], UncomparableObj)):
        fingerprint_list[2] = construct_object_hash(new_obj, deepcopy=True)
    end = time.time()
    profile_dict["representation"] += end - start

    return changed, overwritten



import dill
from collections import defaultdict

from pathlib import Path

from ipykernel.zmqshell import ZMQInteractiveShell

from elastic.core.common.checkpoint_file import CheckpointFile
from elastic.core.graph.graph import DependencyGraph

from elastic.core.io.filesystem_adapter import FilesystemAdapter

# Default checkpoint location if a file path isn't specified.
FILENAME = "./notebook.pickle"


def migrate(graph: DependencyGraph, shell: ZMQInteractiveShell, vss_to_migrate: set, vss_to_recompute: set,
            ces_to_recompute: set, udfs, recomputation_ces, overlapping_vss, filename: str):
    """
        Writes the graph representation of the notebook, migrated variables, and instructions for recomputation as the
        specified file.

        Args:
            graph (DependencyGraph): dependency graph representation of the notebook.
            shell (ZMQInteractiveShell): interactive Jupyter shell storing the state of the current session.
            vss_to_migrate (set): set of VSs to migrate.
            vss_to_recompute (set): set of VSs to recompute.
            ces_to_recompute (set): set of CEs to recompute post-migration.
            filename (str): the location to write the checkpoint to.
            udfs (set): set of user-declared functions.
    """
    # Retrieve variables
    variables = defaultdict(list)
    for vs in vss_to_migrate:
        variables[vs.output_ce].append(vs)

    # construct serialization order list.
    temp_dict = {}
    serialization_order = []
    for (vs1, vs2) in overlapping_vss:
        if vs1 in temp_dict:
            temp_dict[vs1].add(vs2)
        elif vs2 in temp_dict:
            temp_dict[vs2].add(vs1)
        else:
            # create new entry
            new_set = {vs1, vs2}
            temp_dict[vs1] = new_set
            temp_dict[vs2] = new_set

    for vs in vss_to_migrate:
        if vs not in temp_dict:
            temp_dict[vs] = {vs}

    for v in temp_dict.values():
        serialization_order.append(list(v))

    # Construct checkpoint JSON.
    adapter = FilesystemAdapter()
    metadata = CheckpointFile().with_dependency_graph(graph) \
        .with_variables(variables) \
        .with_vss_to_migrate(vss_to_migrate) \
        .with_vss_to_recompute(vss_to_recompute) \
        .with_ces_to_recompute(ces_to_recompute) \
        .with_recomputation_ces(recomputation_ces) \
        .with_serialization_order(serialization_order) \
        .with_udfs(udfs)

    if filename:
        print("Checkpoint saved to:", filename)
        write_path = filename
    else:
        write_path = FILENAME

    with open(Path(write_path), "wb") as output_file:
        dill.dump(metadata, output_file)
        for vs_list in serialization_order:
            obj_list = []
            for vs in vs_list:
                obj_list.append(shell.user_ns[vs.name])
            dill.dump(obj_list, output_file)
    # Write the JSON file to the specified location. Uses the default location if a file path isn't specified.
    #if filename:
    #    print("Checkpoint saved to:", filename)
    #    adapter.write_all(Path(filename), metadata)
    #else:
       # adapter.write_all(Path(FILENAME), metadata)



import json
from typing import Dict
from elastic.core.graph.graph import DependencyGraph

KEY_DEPENDENCY_GRAPH = "dependencyGraph"
KEY_VARIABLES = "variables"
KEY_VSS_TO_MIGRATE = "vss_to_migrate"
KEY_VSS_TO_RECOMPUTE = "vss_to_recompute"
KEY_CES_TO_RECOMPUTE = "ces_to_recompute"
KEY_RECOMPUTATION_CES = "recomputation_ces"
KEY_SERIALIZATION_ORDER = "serialization_order"
KEY_UDFS = "udfs"


class CheckpointFile:
    """
        JSON representation of the notebook checkpoint.
    """
    def __init__(self):
        # Dependency graph representation of the notebook.
        self.dependency_graph = None

        # Migrated variables.
        self.variables = None

        # Active VSs corresponding to migrated variables.
        self.vss_to_migrate = None

        # Variables to recompute post-migration.
        self.vss_to_recompute = None

        # CEs to recompute to restore non-migrated variables (vss_to_recompute).
        self.ces_to_recompute = None

        # CEs to recompute a given CE. For fault tolerance if certain VSs fail
        # to deserialize.
        self.recomputation_ces = None

        # List of objects packed in the pickle file.
        self.serialization_order = None

        # User-declared functions in the session.
        self.udfs = None

    def with_dependency_graph(self, graph: DependencyGraph):
        self.dependency_graph = graph
        return self

    def get_dependency_graph(self):
        return self.dependency_graph

    def with_variables(self, variables: Dict):
        self.variables = variables
        return self

    def get_variables(self):
        return self.variables

    def with_vss_to_migrate(self, vss_to_migrate: set):
        self.vss_to_migrate = vss_to_migrate
        return self

    def get_vss_to_migrate(self):
        return self.vss_to_migrate

    def with_vss_to_recompute(self, vss_to_recompute: set):
        self.vss_to_recompute = vss_to_recompute
        return self

    def get_vss_to_recompute(self):
        return self.vss_to_recompute

    def with_ces_to_recompute(self, ces_to_recompute: set):
        self.ces_to_recompute = ces_to_recompute
        return self

    def get_ces_to_recompute(self):
        return self.ces_to_recompute

    def with_recomputation_ces(self, recomputation_ces: dict):
        self.recomputation_ces = recomputation_ces
        return self

    def get_recomputation_ces(self):
        return self.recomputation_ces

    def with_serialization_order(self, serialization_order: list):
        self.serialization_order = serialization_order
        return self

    def get_serialization_order(self):
        return self.serialization_order

    def with_udfs(self, udfs: set):
        self.udfs = udfs
        return self

    def get_udfs(self):
        return self.udfs

    def to_json_str(self) -> str:
        return json.dumps({
            KEY_DEPENDENCY_GRAPH: self.dependency_graph,
            KEY_VARIABLES: self.variables,
            KEY_VSS_TO_MIGRATE: self.vss_to_migrate,
            KEY_VSS_TO_RECOMPUTE: self.vss_to_recompute,
            KEY_CES_TO_RECOMPUTE: self.ces_to_recompute,
            KEY_RECOMPUTATION_CES: self.recomputation_ces,
            KEY_SERIALIZATION_ORDER: self.serialization_order,
            KEY_UDFS: self.udfs
        })

    @staticmethod
    def from_json(kv: Dict):
        return CheckpointFile().with_dependency_graph(kv[KEY_DEPENDENCY_GRAPH])\
                                  .with_variables(kv[KEY_VARIABLES])\
                                  .with_vss_to_migrate(kv[KEY_VSS_TO_MIGRATE])\
                                  .with_vss_to_recompute(kv[KEY_VSS_TO_RECOMPUTE])\
                                  .with_ces_to_recompute(kv[KEY_CES_TO_RECOMPUTE])\
                                  .with_recomputation_ces(kv[KEY_RECOMPUTATION_CES])\
                                  .with_serialization_order(kv[KEY_SERIALIZATION_ORDER])\
                                  .with_udfs(kv[KEY_UDFS])



import time
from typing import Dict

import numpy as np
from elastic.algorithm.selector import Selector
from elastic.core.graph.graph import DependencyGraph
from elastic.core.io.migrate import migrate
from elastic.core.io.pickle import is_picklable
from elastic.core.common.profile_variable_size import profile_variable_size
from ipykernel.zmqshell import ZMQInteractiveShell

from elastic.core.mutation.object_hash import UnserializableObj


def checkpoint(graph: DependencyGraph, shell: ZMQInteractiveShell, fingerprint_dict: Dict,
               selector: Selector, udfs: set, filename: str, profile_dict, write_log_location=None, notebook_name=None,
               optimizer_name=None):
    """
        Checkpoints the notebook. The optimizer selects the VSs to migrate and recompute and the OEs to recompute, then
        writes the checkpoint as the specified filename.
        Args:
            graph (DependencyGraph): dependency graph representation of the notebook.
            shell (ZMQInteractiveShell): interactive Jupyter shell storing the state of the current session.
            selector (Selector): optimizer for computing the checkpointing configuration.
            udfs (set): set of user-declared functions.
            filename (str): location to write the file to.
            write_log_location (str): location to write component runtimes to. For experimentation only.
            notebook_name (str): notebook name. For experimentation only.
            optimizer_name (str): optimizer name. For experimentation only.
    """
    profile_start = time.time()

    # Retrieve active VSs from the graph. Active VSs are correspond to the latest instances/versions of each variable.
    active_vss = set()
    print("---------------------------")
    print("all variables:")
    # print(graph.variable_snapshots)
    for vs_list in graph.variable_snapshots.values():
        if not vs_list[-1].deleted:
            print(f"name: {vs_list[-1].name}")
            active_vss.add(vs_list[-1])

    # Profile the size of each variable defined in the current session.
    for active_vs in active_vss:
        attr_str = getattr(shell.user_ns[active_vs.name], '__module__', None)
        # Object is unserializable
        if isinstance(fingerprint_dict[active_vs.name][2], UnserializableObj):
            active_vs.size = np.inf

        # Blacklisted object
        elif attr_str and ("dataprep.eda" in attr_str or "bokeh" in attr_str):
            active_vs.size = np.inf

        # Profile size of object.
        else:
            active_vs.size = profile_variable_size(shell.user_ns[active_vs.name])

    # Check for pairwise variable intersections. Variables sharing underlying data must be migrated or recomputed
    # together.
    overlapping_vss = []
    for active_vs1 in active_vss:
        for active_vs2 in active_vss:
            if active_vs1 != active_vs2 and fingerprint_dict[active_vs1.name][1].intersection(
                    fingerprint_dict[active_vs2.name][1]):
                overlapping_vss.append((active_vs1, active_vs2))

    profile_end = time.time()
    if write_log_location:
        with open(write_log_location + '/output_' + notebook_name + '_' + optimizer_name + '.txt', 'a') as f:
            f.write('overlappings - ' + repr(len(overlapping_vss)) + '\n')
            f.write('Profile stage took - ' + repr(profile_start - profile_end) + " seconds" + '\n')
            f.write('Idgraph stage took - ' + repr(profile_dict["idgraph"]) + " seconds" + '\n')
            f.write('Representation stage took - ' + repr(profile_dict["representation"]) + " seconds" + '\n')

    optimize_start = time.time()
    # Initialize the optimizer.
    add_start = time.time()
    selector.dependency_graph = graph
    selector.active_vss = active_vss
    selector.overlapping_vss = overlapping_vss
    add_end = time.time()

    # Use the optimizer to compute the checkpointing configuration.
    opt_start = time.time()
    vss_to_migrate, ces_to_recompute = selector.select_vss(write_log_location, notebook_name, optimizer_name)
    opt_end = time.time()
    print("---------------------------")
    print("variables to migrate:")
    for vs in vss_to_migrate:
        print(f"name: {vs.name}, size: {vs.size}")

    difference_start = time.time()
    vss_to_recompute = active_vss - vss_to_migrate
    difference_end = time.time()

    print("---------------------------")
    print("variables to recompute:")
    for vs in vss_to_recompute:
        print(f"name: {vs.name}, size: {vs.size}")
    print([vs.name for vs in vss_to_recompute])

    print("---------------------------")
    print("cells to recompute:")
    for ce in ces_to_recompute:
        print(f"cell num: {ce.cell_num}, cell runtime: {ce.cell_runtime}")
    print(sorted([ce.cell_num + 1 for ce in ces_to_recompute]))

    optimize_end = time.time()
    if write_log_location:
        with open(write_log_location + '/output_' + notebook_name + '_' + optimizer_name + '.txt', 'a') as f:
            f.write('Optimize stage took - ' + repr(optimize_end - optimize_start) + " seconds" + '\n')
            f.write('  Add stage took - ' + repr(add_end - add_start) + " seconds" + '\n')
            f.write('  Opt stage took - ' + repr(opt_end - opt_start) + " seconds" + '\n')
            f.write('  Diff stage took - ' + repr(difference_end - difference_start) + " seconds" + '\n')

    # Store the notebook checkpoint to the specified location.
    migrate_start = time.time()
    migrate_success = True
    migrate(graph, shell, vss_to_migrate, vss_to_recompute, ces_to_recompute, udfs, selector.recomputation_ces, selector.overlapping_vss, filename)
    migrate_end = time.time()

    if write_log_location:
        with open(write_log_location + '/output_' + notebook_name + '_' + optimizer_name + '.txt', 'a') as f:
            f.write('Migrate stage took - ' + repr(migrate_end - migrate_start) + " seconds" + '\n')
    return migrate_success



import numpy as np

from elastic.algorithm.selector import Selector


class MigrateAllBaseline(Selector):
    """
        Migrates all active VSs.
    """
    def __init__(self, migration_speed_bps=1):
        super().__init__(migration_speed_bps)

    def select_vss(self) -> set:
        vss_to_migrate = set()
        for vs in self.active_vss:
            if vs.size < np.inf:
                vss_to_migrate.add(vs)
        return vss_to_migrate


class RecomputeAllBaseline(Selector):
    """
        Recomputes all active VSs.
    """
    def __init__(self, migration_speed_bps=1):
        super().__init__(migration_speed_bps)

    def select_vss(self) -> set:
        return set()



from __future__ import print_function

import sys, traceback
import time
import types
from os.path import dirname
from pympler import asizeof

from IPython import get_ipython
from IPython.core.interactiveshell import InteractiveShell
from elastic.algorithm.selector import OptimizerType

from elastic.core.common.profile_migration_speed import profile_migration_speed
from elastic.core.common.profile_graph_size import profile_graph_size
from elastic.core.notebook.checkpoint import checkpoint
from elastic.core.notebook.update_graph import update_graph
from elastic.core.notebook.find_input_vars import find_input_vars
from elastic.core.notebook.find_output_vars import find_created_deleted_vars
from elastic.core.mutation.fingerprint import construct_fingerprint, compare_fingerprint
from elastic.core.mutation.object_hash import UnserializableObj
from elastic.core.notebook.restore_notebook import restore_notebook

from elastic.core.graph.graph import DependencyGraph
from elastic.core.io.recover import resume

from elastic.algorithm.optimizer_exact import OptimizerExact
from elastic.algorithm.baseline import RecomputeAllBaseline, MigrateAllBaseline


class ElasticNotebook():
    """
        Magics class for Elastic Notebook. Enable this in the notebook by running '%load_ext ElasticNotebook'.
        Enables efficient checkpointing of intermediate notebook state via balancing migration and recomputation
        costs.
    """

    def __init__(self, shell: InteractiveShell):
        self.shell = shell

        # Initialize the dependency graph for capturing notebook state.
        self.dependency_graph = DependencyGraph()

        # Migration properties.
        self.migration_speed_bps = 100000
        self.alpha = 1
        self.selector = OptimizerExact(migration_speed_bps=self.migration_speed_bps)

        # Dictionary of object fingerprints. For detecting modified references.
        self.fingerprint_dict = {}

        # Set of user-declared functions.
        self.udfs = set()

        # Flag if migration speed has been manually set. In this case, skip profiling of migration speed at checkpoint
        # time.
        self.manual_migration_speed = False

        # Location to log runtimes to. For experiments only.
        self.write_log_location = None

        # Strings for determining log filename. For experiments only.
        self.optimizer_name = ""
        self.notebook_name = ""

        # Total elapsed time spent inferring cell inputs and outputs.
        # For measuring overhead.
        self.total_recordevent_time = 0

        # Dict for recording overhead of profiling operations.
        self.profile_dict = {"idgraph": 0, "representation": 0}

        # デバッグフラグ
        self.debug = False

    def set_debug(self, debug):
        self.debug = debug

    def record_event(self, cell):
        if self.debug:
            print("Recording event...")

        pre_execution = set(self.shell.user_ns.keys())

        # Create id trees for output variables
        for var in self.dependency_graph.variable_snapshots.keys():
            if var not in self.fingerprint_dict and var in self.shell.user_ns:
                self.fingerprint_dict[var] = construct_fingerprint(self.shell.user_ns[var], self.profile_dict)

        # Find input variables (variables potentially accessed) of the cell.
        input_variables, function_defs = find_input_vars(cell, set(self.dependency_graph.variable_snapshots.keys()),
                                                         self.shell, self.udfs)
        # Union of ID graphs of input variables. For detecting modifications to unserializable variables.
        input_variables_id_graph_union = set()
        for var in input_variables:
            if var in self.fingerprint_dict:
                input_variables_id_graph_union = input_variables_id_graph_union.union(self.fingerprint_dict[var][1])

        # Run the cell.
        start_time = time.time()
        try:
            cell_output = get_ipython().run_cell(cell)
            cell_output.raise_error()
            # traceback_list = []
        except:
            pass
            # _, _, tb = sys.exc_info()
            # traceback_list = traceback.extract_tb(tb).format()
        cell_runtime = time.time() - start_time
        post_execution = set(self.shell.user_ns.keys())
        infer_start = time.time()

        # Find created and deleted variables by computing difference between namespace pre and post execution.
        created_variables, deleted_variables = find_created_deleted_vars(pre_execution, post_execution)

        # Remove stored ID graphs for deleted variables.
        for var in deleted_variables:
            del self.fingerprint_dict[var]
            if var in self.udfs:
                self.udfs.remove(var)

        # Find modified variables by comparing ID graphs and object hashes.
        modified_variables = set()
        for k, v in self.fingerprint_dict.items():
            changed, overwritten = compare_fingerprint(self.fingerprint_dict[k], self.shell.user_ns[k],
                                                       self.profile_dict, input_variables_id_graph_union)
            if changed:
                modified_variables.add(k)

            # In the case of non-overwrite modification, the variable is additionally considered as accessed.
            if changed and not overwritten:
                input_variables.add(k)

            # A user defined function has been overwritten.
            elif overwritten and k in self.udfs:
                self.udfs.remove(k)

            # Select unserializable variables are assumed to be modified if accessed.
            if not changed and not overwritten and isinstance(self.fingerprint_dict[k][2], UnserializableObj):
                if self.fingerprint_dict[k][1].intersection(input_variables_id_graph_union):
                    modified_variables.add(k)

        # Create ID graphs for output variables
        for var in created_variables:
            self.fingerprint_dict[var] = construct_fingerprint(self.shell.user_ns[var], self.profile_dict)

        # Record newly defined UDFs
        for udf in function_defs:
            if udf in self.shell.user_ns and isinstance(self.shell.user_ns[udf], types.FunctionType):
                self.udfs.add(udf)

        # Update the dependency graph.
        update_graph(cell, cell_runtime, start_time, input_variables, created_variables.union(modified_variables),
                     deleted_variables, self.dependency_graph)

        # Update total recordevent time tally.
        infer_end = time.time()
        self.total_recordevent_time += infer_end - infer_start

    def set_migration_speed(self, migration_speed):
        try:
            if float(migration_speed) > 0:
                self.migration_speed_bps = float(migration_speed)
                self.manual_migration_speed = True
            else:
                pass
                # print("Migration speed is not positive.")
        except ValueError:
            print("Migration speed is not a number.")
        self.selector.migration_speed_bps = self.migration_speed_bps

    def set_optimizer(self, optimizer):
        self.optimizer_name = optimizer

        if optimizer == OptimizerType.EXACT.value:
            self.selector = OptimizerExact(self.migration_speed_bps)
            self.alpha = 1
        elif optimizer == OptimizerType.EXACT_C.value:
            self.selector = OptimizerExact(self.migration_speed_bps)
            self.alpha = 20
        elif optimizer == OptimizerType.EXACT_R.value:
            self.selector = OptimizerExact(self.migration_speed_bps)
            self.alpha = 0.05
        elif optimizer == OptimizerType.MIGRATE_ALL.value:
            self.selector = MigrateAllBaseline(self.migration_speed_bps)
        elif optimizer == OptimizerType.RECOMPUTE_ALL.value:
            self.selector = RecomputeAllBaseline(self.migration_speed_bps)

    def set_write_log_location(self, filename):
        self.write_log_location = filename

    def set_notebook_name(self, name):
        self.notebook_name = name

    def checkpoint(self, filename):
        if self.debug:
            print("Checkpointing...")

        # Write overhead metrics to file (for experiments).
        if self.write_log_location:
            with open(self.write_log_location + '/output_' + self.notebook_name + '_' +
                      self.optimizer_name + '.txt', 'a') as f:
                f.write('comparison overhead - ' + repr(asizeof.asizeof(self.dependency_graph) +
                                                        asizeof.asizeof(self.fingerprint_dict)) + ' bytes' + '\n')
                f.write('notebook overhead - ' + repr(asizeof.asizeof(self.shell.user_ns)) + ' bytes' + '\n')
                f.write('Dependency graph storage overhead - ' + repr(profile_graph_size(self.dependency_graph)) +
                        " bytes" + '\n')
                f.write('Cell monitoring overhead - ' + repr(self.total_recordevent_time) + " seconds" + '\n')

        # Profile the migration speed to filename.
        if not self.manual_migration_speed:
            self.migration_speed_bps = profile_migration_speed(dirname(filename), alpha=self.alpha)
            self.selector.migration_speed_bps = self.migration_speed_bps

        # Checkpoint the notebook.
        checkpoint(self.dependency_graph, self.shell, self.fingerprint_dict, self.selector, self.udfs,
                   filename, self.profile_dict, self.write_log_location, self.notebook_name, self.optimizer_name)

    def load_checkpoint(self, filename):
        if self.debug:
            print("Loading checkpoint...")

        # start_time = time.time()
        self.dependency_graph, variables, vss_to_migrate, vss_to_recompute, oes_to_recompute, self.udfs = \
            resume(filename, self.write_log_location, self.notebook_name, self.optimizer_name)
        
        # Recompute missing VSs and redeclare variables into the kernel.
        restore_notebook(self.dependency_graph, self.shell, variables, oes_to_recompute, self.write_log_location,
                         self.notebook_name, self.optimizer_name)
        # print("Checkpoint load time:", time.time() - start_time)



from elastic.core.graph.graph import DependencyGraph
from elastic.core.common.profile_variable_size import profile_variable_size
import sys


def profile_graph_size(graph: DependencyGraph):
    """
       Profiles the in-memory size of a dependency graph of Elastic Notebook.
       For experiments only.
       Args:
           graph (DependencyGraph): dependency graph to profile.
    """

    # Tally fields in graph
    total_size = sys.getsizeof(graph)
    total_size += sys.getsizeof(graph.cell_executions)
    total_size += profile_variable_size(graph.variable_snapshots)

    # Tally fields in operation events
    for oe in graph.cell_executions:
        total_size += sys.getsizeof(oe.cell_num)
        total_size += sys.getsizeof(oe.cell)
        total_size += sys.getsizeof(oe.cell_runtime)
        total_size += sys.getsizeof(oe.start_time)
        total_size += sys.getsizeof(oe.src_vss)
        total_size += sys.getsizeof(oe.dst_vss)

    # Tally fields in nodes
    for vs_list in graph.variable_snapshots.values():
        for vs in vs_list:
            total_size += sys.getsizeof(vs.name)
            total_size += sys.getsizeof(vs.version)
            total_size += sys.getsizeof(vs.deleted)
            total_size += sys.getsizeof(vs.output_ce)
            total_size += sys.getsizeof(vs.input_ces)
            total_size += sys.getsizeof(vs.size)

    return total_size



from enum import Enum


class OptimizerType(Enum):
    """
        String representations of all implemented optimizers.
    """
    EXACT = "exact"
    EXACT_C = "exact_c"
    EXACT_R = "exact_r"
    MIGRATE_ALL = "migrate_all"
    RECOMPUTE_ALL = "recompute_all"


class Selector:
    """
        The `Selector` class provides interfaces to pick a subset of active VSs to migrate based on
            various heuristics and algorithms.
    """

    def __init__(self, migration_speed_bps=1):
        """
            Creates a Selector instance with a migration speed estimate. The dependency graph and active VS fields
            must be populated prior to calling select_vss.
        """
        self.dependency_graph = None
        self.active_vss = None
        self.overlapping_vss = None
        self.migration_speed_bps = migration_speed_bps

    def select_vss(self) -> set:
        """
            Classes that inherit from the `Selector` class (such as `Optimizer` and various baselines)
                should override `select_vss`.

            Returns:
                set(VariableSnapshot): a subset of active VSs selected to migrate based on various heuristics and
                algorithms.
        """
        raise NotImplementedError()



from typing import Tuple
import ast
import inspect
from collections import deque
from ipykernel.zmqshell import ZMQInteractiveShell

PRIMITIVES = {int, bool, str, float}


# Node visitor for finding input variables.
class Visitor(ast.NodeVisitor):
    def __init__(self, shell, shell_udfs):
        # Whether we are currently in local scope.
        self.is_local = False

        # Functions declared in 
        self.functiondefs = set()
        self.udfcalls = set()
        self.loads = set()
        self.globals = set()
        self.shell = shell
        self.udfs = shell_udfs

    def generic_visit(self, node):
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Name(self, node):
        if isinstance(node.ctx, ast.Load):
            # Only add as input if variable exists in current scope.
            if not (self.is_local and node.id not in self.globals and node.id in self.shell.user_ns and
                    type(self.shell.user_ns[node.id]) in PRIMITIVES):
                self.loads.add(node.id)
        ast.NodeVisitor.generic_visit(self, node)

    def visit_AugAssign(self, node):
        # Only add as input if variable exists in current scope.
        if isinstance(node.target, ast.Name):
            if not (self.is_local and node.target.id not in self.globals and node.target.id in self.shell.user_ns and
                    type(self.shell.user_ns[node.target.id]) in PRIMITIVES):
                self.loads.add(node.target.id)
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Global(self, node):
        for name in node.names:
            self.globals.add(name)
        ast.NodeVisitor.generic_visit(self, node)

    def visit_Call(self, node):
        if isinstance(node.func, ast.Name):
            self.udfcalls.add(node.func.id)
        ast.NodeVisitor.generic_visit(self, node)

    def visit_FunctionDef(self, node):
        # Only add as input if variable exists in current scope
        self.is_local = True
        self.functiondefs.add(node.name)
        ast.NodeVisitor.generic_visit(self, node)
        self.is_local = False


def find_input_vars(cell: str, existing_variables: set, shell: ZMQInteractiveShell, shell_udfs: set) -> Tuple[
    set, dict]:
    """
        Capture the input variables of the cell via AST analysis.
        Args:
            cell (str): Raw cell cell.
            existing_variables (set): Set of user-defined variables in the current session.
            shell (ZMQInteractiveShell): Shell of current session. For inferring variable types.
            shell_udfs (set): Set of user-declared functions in the shell.
    """
    # Initialize AST walker.
    v1 = Visitor(shell=shell, shell_udfs=shell_udfs)

    # Parse the cell code.
    v1.visit(ast.parse(cell))

    # Find top-level input variables and function declarations.
    input_variables = v1.loads
    function_defs = v1.functiondefs

    # Recurse into accessed UDFs.
    udf_calls = deque()
    udf_calls_visited = set()

    for udf in v1.udfcalls:
        if udf not in udf_calls_visited and udf in shell_udfs:
            udf_calls.append(udf)
            udf_calls_visited.add(udf)

    while udf_calls:
        # Visit the next nested UDF call.
        v_nested = Visitor(shell=shell, shell_udfs=shell_udfs)
        udf = udf_calls.popleft()
        v_nested.visit(ast.parse(inspect.getsource(shell.user_ns[udf])))

        # Update input variables and function definitions
        input_variables = input_variables.union(v_nested.loads)
        function_defs = function_defs.union(v_nested.functiondefs)
        for udf in v_nested.udfcalls:
            if udf not in udf_calls_visited and udf in shell_udfs:
                udf_calls.append(udf)
                udf_calls_visited.add(udf)

    # A variable is an input only if it is in the shell before cell execution.
    input_variables = input_variables.intersection(existing_variables)

    return input_variables, function_defs



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2021-2022 University of Illinois
import time
import dill
from collections import defaultdict
from pathlib import Path

from elastic.core.io.filesystem_adapter import FilesystemAdapter

from elastic.core.io.migrate import FILENAME


def resume(filename: str, write_log_location=None, notebook_name=None, optimizer_name=None):
    """
    Reads the file at `filename` and unpacks the graph representation of the notebook, migrated variables, and
    instructions for recomputation.

    Args:
        filename (str): Location of the checkpoint file.

        write_log_location (str): location to write file read runtime to. For experimentation only.
        notebook_name (str): notebook name. For experimentation only.
        optimizer_name (str): optimizer name. For experimentation only.
    """

    # Reads from the default location if a file path isn't specified.
    adapter = FilesystemAdapter()

    load_start = time.time()
    if filename:
        read_path = filename
    else:
        read_path = FILENAME

    variables = defaultdict(list)

    with open(Path(read_path), "rb") as output_file:
        metadata = dill.load(output_file)
        print(metadata.get_serialization_order())
        for vs_list in metadata.get_serialization_order():
            try:
                print("trying:", [vs.name for vs in vs_list])
                obj_list = dill.load(output_file)
                for i in range(len(vs_list)):
                    print("me: ", vs_list[i].output_ce.cell_num)
                    variables[vs_list[i].output_ce.cell_num].append((vs_list[i], obj_list[i]))
            except:
                print("oops")
                #unpickling failed. Rerun cells to retrieve variable(s).
                for vs in vs_list:
                    metadata.ces_to_recompute = metadata.ces_to_recompute.union(
                            metadata.recomputation_ces(vs.output_ce))


    if filename:
        metadata = adapter.read_all(Path(filename))
    else:
        metadata = adapter.read_all(Path(FILENAME))
    load_end = time.time()

    if write_log_location:
        with open(write_log_location + '/output_' + notebook_name + '_' + optimizer_name + '.txt', 'a') as f:
            f.write('Reload stage took - ' + repr(load_end - load_start) + " seconds\n")

    return metadata.get_dependency_graph(), variables, metadata.get_vss_to_migrate(), \
        metadata.get_vss_to_recompute(), metadata.get_ces_to_recompute(), metadata.get_udfs()



import os, sys, time
import numpy as np
import pickle


def profile_migration_speed(dirname: str, alpha=1) -> float:
    """
        The migration speed is the sum of read and write speed (since we are writing the state to disk, then
        reading from disk to restore the notebook). The function should ideally be fast (<1 sec).
        Args:
            dirname: Location to profile.
    """
    filecount = 1
    max_time = 0.8
    testing_dir = os.path.join(dirname, "measure_speed")
    os.system("rm -rf {} && mkdir {}".format(testing_dir, testing_dir))
    total_bytes = 0

    start_time = time.time()

    total_read_time = 0
    total_write_time = 0
    for i in range(filecount):
        # write_array_large = np.random.rand(10000, 10000)
        write_array_large = np.random.rand(500, 500)
        write_array_small = np.random.rand(100, 100)
        total_bytes += sys.getsizeof(write_array_large)
        total_bytes -= sys.getsizeof(write_array_small)

        write_start = time.time()
        out_file = open(os.path.join(testing_dir, str(i) + "large"), "wb")
        pickle.dump(write_array_large, out_file)
        out_file.close()
        total_write_time += time.time() - write_start

        read_start = time.time()
        in_file = open(os.path.join(testing_dir, str(i)) + "large", "rb")
        in_file.close()
        total_read_time += time.time() - read_start

        write_start = time.time()
        out_file = open(os.path.join(testing_dir, str(i) + "small"), "wb")
        pickle.dump(write_array_small, out_file)
        out_file.close()
        total_write_time -= time.time() - write_start

        read_start = time.time()
        in_file = open(os.path.join(testing_dir, str(i) + "small"), "rb")
        in_file.close()
        total_read_time -= time.time() - read_start

        if time.time() - start_time > max_time:
            break

    os.system("rm -rf {}".format(testing_dir))

    migration_speed_bps = total_bytes / (total_read_time + total_write_time * alpha)
    print("migration speed (bps) (total_bytes / (total_read_time + total_write_time * alpha)):", migration_speed_bps)
    print("|- total bytes:", total_bytes)
    print("|- total read time:", total_read_time)
    print("|- total write time:", total_write_time)
    print("|- alpha:", alpha)
    return migration_speed_bps



import logging
from collections import defaultdict
from typing import List

from elastic.core.graph.variable_snapshot import VariableSnapshot
from elastic.core.graph.cell_execution import CellExecution


class DependencyGraph:
    """
        A dependency graph is a snapshot of the history of a notebook instance.
        Nodesets and operation events are the nodes and edges of the dependency graph.
    """
    def __init__(self):
        """
            Create a new dependency graph. Called when the magic extension of elastic notebook is loaded with %load_ext.
        """
        # Cell executions.
        self.cell_executions = []

        # Dict of variable snapshots.
        # Keys are variable names, while values are lists of the actual VSs.
        # i.e. {"x": [(x, 1), (x, 2)], "y": [(y, 1), (y, 2), (y, 3)]}
        self.variable_snapshots = defaultdict(list)

    def create_variable_snapshot(self, variable_name: str, deleted: bool) -> VariableSnapshot:
        """
            Creates a new variable snapshot for a given variable.
            Args:
                variable_name (str): variable_name
                deleted (bool): Whether this VS is created for the deletion of a variable, i.e. 'del x'.
        """

        # Assign a version number to the VS.
        if variable_name in self.variable_snapshots:
            version = len(self.variable_snapshots[variable_name])
        else:
            version = 0

        # Create a new VS instance and store it in the graph.
        vs = VariableSnapshot(variable_name, version, deleted)
        self.variable_snapshots[variable_name].append(vs)
        return vs

    def add_cell_execution(self, cell, cell_runtime: float, start_time: float, src_vss: List, dst_vss: List):
        """
            Create a cell execution from captured metrics.
            Args:
                cell (str): Raw cell cell.
                cell_runtime (float): Cell runtime.
                start_time (time): Time of start of cell execution. Note that this is different from when the cell was
                    queued.
                src_vss (List): List containing input VSs of the cell execution.
                dst_vss (List): List containing output VSs of the cell execution.
        """

        # Create a cell execution.
        ce = CellExecution(len(self.cell_executions), cell, cell_runtime, start_time, src_vss, dst_vss)

        # Add the newly created cell execution to the graph.
        self.cell_executions.append(ce)

        # Set the newly created cell execution as dependent on its input variable snapshots.
        for src_vs in src_vss:
            src_vs.input_ces.append(ce)

        # Set the newly created cell execution as the parent of its output variable snapshots.
        for dst_vs in dst_vss:
            dst_vs.output_ce = ce



from inspect import isclass
from types import ModuleType

PRIMITIVES = {type(None), int, float, bool, str}
ITERABLES = {tuple, set, list}
OBJECT_FILTER_FUNC = lambda x: not x[0].startswith("_") and not is_primitive(type(x[1]))
ITERABLE_FILTER_FUNC = lambda x: not is_primitive(x)


class IdGraphNode:
    """
        The IdGraph is used to model the reachable objects (their ids and types) from a given object.
    """

    def __init__(self, obj_id, obj_type, child_nodes):
        """
            Args:
                obj_id: id (memory address) of the object.
                obj_type: type of the object.
                child_nodes: other objects reachable from this object (i.e. class attributes, list/set members)
        """
        self.obj_id = obj_id
        self.obj_type = obj_type
        self.child_nodes = child_nodes


def is_primitive(obj_type):
    return obj_type in PRIMITIVES


def is_root_equals(node1, node2):
    """
        Compare only the root notes of 2 ID graphs.
    """
    if node1 is None and node2 is None:
        return True
    elif node1 is None or node2 is None:
        return False
    return node1.obj_id == node2.obj_id and node1.obj_type == node2.obj_type


def is_structure_equals_helper(node1, node2, visited):
    """
        BFS helper for recursively comparing ID graphs.
    """
    if node1.obj_id != node2.obj_id or node1.obj_type != node2.obj_type or len(node1.child_nodes) != len(
            node2.child_nodes):
        return False
    visited.add(node1)

    for i in range(len(node1.child_nodes)):
        if node1.child_nodes[i] in visited:
            continue
        if not is_structure_equals_helper(node1.child_nodes[i], node2.child_nodes[i], visited):
            return False

    return True


def is_structure_equals(node1, node2):
    """
        Compare 2 ID graphs for equality.
    """
    if node1 is None and node2 is None:
        return True
    elif node1 is None or node2 is None:
        return False
    return is_structure_equals_helper(node1, node2, set())


def construct_id_graph_node(obj, visited):
    """
        Helper function for constructing an ID graph. Constructs an ID graph node and recurses into reachable objects
        in BFS fashion.
        Args:
            obj: object.
            visited (set): set of visited objects (for handling cyclic references).
    """

    if obj is None:
        return None

    # Construct ID graph node.
    T = type(obj)
    child_list = []
    id_graph_node = IdGraphNode(id(obj), T.__name__, child_list)

    visited[id(obj)] = id_graph_node

    # obj is Iterable: recurse into items
    if T in ITERABLES:
        for child_obj in obj:
            if is_primitive(type(child_obj)):
                continue
            if id(child_obj) in visited:
                child_list.append(visited[id(child_obj)])
            else:
                child_list.append(construct_id_graph_node(child_obj, visited))

    # obj is object instance: recurse into attributes
    # Special note for sets: this works as the order of iterating through sets in the same session is deterministic.
    elif hasattr(obj, '__dict__') and not isinstance(obj, ModuleType) and not isclass(obj):
        for k, v in filter(OBJECT_FILTER_FUNC, vars(obj).items()):
            if id(v) in visited:
                child_list.append(visited[id(v)])
            else:
                child_list.append(construct_id_graph_node(v, visited))

    # obj is dictionary: recurse into both keys and values
    elif T is dict:
        for k, v in obj.items():
            if is_primitive(type(k)):
                continue
            if id(k) in visited:
                child_list.append(visited[id(k)])
            else:
                child_list.append(construct_id_graph_node(k, visited))

            if is_primitive(type(v)):
                continue
            if id(v) in visited:
                child_list.append(visited[id(v)])
            else:
                child_list.append(construct_id_graph_node(v, visited))

    return id_graph_node


def construct_id_graph(obj):
    """
        Construct the ID graph for an object.
        Returns:
            graph_bytestring (bytestring): the serialized bytestring of the ID graph.
            visited: a set of all IDs of reachable objects in the graph. For checking object overlaps.
    """
    visited = {}
    graph = construct_id_graph_node(obj, visited)
    return graph, set(visited.keys())



import sys
import types


def get_total_size(data):
    """
        Compute the estimated total size of a variable.
    """
    def get_memory_size(obj, is_initialize, visited):
        # same memory space should be calculated only once
        obj_id = id(obj)
        if obj_id in visited:
            return 0
        visited.add(obj_id)

        try:
            total_size = sys.getsizeof(obj)
        except Exception:
            total_size = float("inf")

        obj_type = type(obj)
        if obj_type in [int, float, str, bool, type(None)]:
            # if the original obj is not primitive, then the size is already included
            if not is_initialize:
                return 0
        else:
            if obj_type in [list, tuple, set]:
                for e in obj:
                    total_size = total_size + get_memory_size(e, False, visited)
            elif obj_type is dict:
                for k, v in obj.items():
                    total_size = total_size + get_memory_size(k, False, visited)
                    total_size = total_size + get_memory_size(v, False, visited)
            # function, method, class
            elif obj_type in [types.FunctionType, types.MethodType, types.BuiltinFunctionType, types.ModuleType] \
                    or isinstance(obj, type):  # True if obj is a class
                pass
            # custom class instance
            elif isinstance(type(obj), type):
                # if obj has no builtin size and has additional pointers
                # if obj has builtin size, all the additional memory space is already added
                if not hasattr(obj, "__sizeof__") and hasattr(obj, "__dict__"):
                    for (k, v) in getattr(obj, "__dict__").items():
                        total_size = total_size + get_memory_size(k, False, visited)
                        total_size = total_size + get_memory_size(v, False, visited)
            else:
                raise NotImplementedError("Not handled", obj)
        return total_size

    return get_memory_size(data, True, set())


def profile_variable_size(x) -> int:
    """
        Profiles the size of variable x. Notably, this should recursively find the size of lists, sets and dictionaries.
        Args:
            x: The variable to profile.
    """
    return get_total_size(x)



def find_created_deleted_vars(pre_execution, post_execution):
    """
        Find created and deleted variables through computing a difference of the user namespace pre and post execution.
    """
    created_variables = set()
    deleted_variables = set()

    # New variables
    for varname in post_execution.difference(pre_execution):
        if not varname.startswith('_'):
            created_variables.add(varname)

    # Deleted variables
    for varname in pre_execution.difference(post_execution):
        if not varname.startswith('_'):
            deleted_variables.add(varname)

    return created_variables, deleted_variables



from pathlib import Path
from elastic.core.io.adapter import Adapter
import dill
import gc


class FilesystemAdapter(Adapter):
    def __init__(self):
        super().__init__()

    def read_all(self, path: Path):
        """
            The following (read then decode) is faster vs. directly returning dill.load when network speed is low.
        """
        gc.disable()
        contents_bytestring = open(path, "rb").read()
        contents = dill.loads(contents_bytestring)
        gc.enable()
        return contents

    def create(self, path: Path):
        path.touch()

    def write_all(self, path: Path, buf):
        dill.dump(buf, open(path, "wb"))

    def remove(self, path: Path):
        path.unlink()



from elastic.core.graph.graph import DependencyGraph


def update_graph(cell: str, cell_runtime: float, start_time: float, input_variables: set,
                 created_and_modified_variables: set, deleted_variables: set, graph: DependencyGraph):
    """
        Updates the graph according to the newly executed cell and its input and output variables.
        Args:
             cell (str): Raw cell cell.
             cell_runtime (float): Cell runtime.
             start_time (time): Time of start of cell execution. Note that this is different from when the cell was
                 queued.
             input_variables (set): Set of input variables of the cell.
             created_and_modified_variables (set): set of created and modified variables.
             deleted_variables (set): set of deleted variables
             graph (DependencyGraph): Dependency graph representation of the notebook.
    """

    # Retrieve input variable snapshots
    input_vss = set(graph.variable_snapshots[variable][-1] for variable in input_variables)

    # Create output variable snapshots
    output_vss_create = set(graph.create_variable_snapshot(k, False) for k in created_and_modified_variables)
    output_vss_delete = set(graph.create_variable_snapshot(k, True) for k in deleted_variables)

    # Add the newly created OE to the graph.
    graph.add_cell_execution(cell, cell_runtime, start_time, input_vss, output_vss_create.union(output_vss_delete))



import time

from IPython import get_ipython
from IPython.utils.capture import capture_output
from ipykernel.zmqshell import ZMQInteractiveShell

from elastic.core.graph.graph import DependencyGraph


def restore_notebook(graph: DependencyGraph, shell: ZMQInteractiveShell,
                     variables: dict, ces_to_recompute: set, write_log_location=None, notebook_name=None,
                     optimizer_name=None):
    """
        Restores the notebook. Declares variables back into the kernel and recomputes the CEs to restore non-migrated
        variables.
        Args:
            graph (DependencyGraph): dependency graph representation of the notebook.
            shell (ZMQInteractiveShell): interactive Jupyter shell storing the state of the current session.
            variables (Dict): Mapping from OEs to lists of variables defined in those OEs.
            oes_to_recompute (set): OEs to recompute to restore non-migrated variables.
            write_log_location (str): location to write component runtimes to. For experimentation only.
            notebook_name (str): notebook name. For experimentation only.
            optimizer_name (str): optimizer name. For experimentation only.
    """

    # Recompute OEs following the order they were executed in.
    recompute_start = time.time()
    for ce in graph.cell_executions:
        if ce in ces_to_recompute:
            # Rerun cell code; suppress stdout when rerunning.
            print("Rerunning cell", ce.cell_num + 1)
            cell_capture = capture_output(stdout=True, stderr=True, display=True)
            try:
                with cell_capture:
                    get_ipython().run_cell(ce.cell)
            except Exception as e:
                raise e
        
        # Define output variables in the CE.
        for pair in variables[ce.cell_num]:
            print("Declaring variable", pair[0].name)
            shell.user_ns[pair[0].name] = pair[1]
    
    recompute_end = time.time()

    if write_log_location:
        with open(write_log_location + '/output_' + notebook_name + '_' + optimizer_name + '.txt', 'a') as f:
            f.write('Recompute stage took - ' + repr(recompute_end - recompute_start) + " seconds" + '\n')




from pathlib import Path


class Adapter:
    def __init__(self):
        pass

    def read_all(self, path: Path):
        raise NotImplementedError()

    def create(self, path: Path):
        raise NotImplementedError()

    def write_all(self, path: Path, buf):
        raise NotImplementedError()

    def remove(self, path: Path):
        raise NotImplementedError()



class VariableSnapshot:
    """
        A variable snapshot in the dependency graph corresponds to a version of a variable.
        I.e. if variable 'x' has been assigned 3 times (x = 1, x = 2, x = 3), then 'x' will have 3 corresponding
        variable snapshots.
    """
    def __init__(self, name: str, version: int, deleted: bool):
        """
            Create a variable snapshot from variable properties.
            Args:
                name (str): Variable name.
                version (int): The nth update to the corresponding variable.
                deleted (bool): Whether this VS is created for the deletion of a variable, i.e. 'del x'.
        """
        self.name = name
        self.version = version

        # Whether this VS corresponds to a deleted variable.
        # i.e. if this VS was created for 'del x' we set this to true so this variable is explicitly not considered
        # for migration.
        self.deleted = deleted

        # Cell executions accessing this variable snapshot (i.e. require this variable snapshot to run).
        self.input_ces = []

        # The unique cell execution creating this variable snapshot.
        self.output_ce = None

        # Size of variable pointed to by VS; estimated at migration time.
        self.size = 0





#/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2021-2022 University of Illinois

import dill
import pickle

import pandas as pd
# import polars as pl
import inspect, os
from scipy import sparse
import time
import seaborn
import networkx
import matplotlib.pyplot as plt
import types
import mmap
import hashlib

def is_picklable_fast(obj):
    if type(obj) in {types.GeneratorType, mmap.mmap, hashlib.sha256}:
        return False
    return True

def is_picklable(obj):
    """
        Checks whether an object is pickleable.
    """
    if is_exception(obj) or inspect.ismodule(obj):
        return True
    try:
        # This function can crash.
        return _is_picklable_dill(obj)
    except Exception:
        try:
            # Double check with function from pickle module
            return _is_picklable_raw(obj)
        except Exception:
            return False


def is_exception(obj):
    """
        List of objects which _is_picklable_dill returns false (or crashes) but are picklable.
    """
    if hasattr(obj, '__module__') and getattr(obj, '__module__', None).split(".")[0] in {plt.__name__, seaborn.__name__, networkx.__name__, pd.__name__}:
        return True
    exceptions = [pd.core.frame.DataFrame, sparse.csr.csr_matrix]
    return type(obj) in exceptions


def _is_picklable_raw(obj):
    try:
        # dumps can be slow for large objects that can be pickled
        pickle.dumps(obj)
    except Exception:
        return False
    return True


def _is_picklable_dill(obj):
    # compared to _is_picklable_raw, this may be faster
    # however, dill's correctness is worrying because
    #   it currently considers Pandas DataFrame as not
    #   picklable, which is not true
    return dill.pickles(obj)
